/*
 * kgraft_patch_bsc1197211
 *
 * Fix for CVE-2021-39713, bsc#1197211
 *
 *  Upstream commits:
 *  6f99528e9797 ("net: core: netlink: add helper refcount dec and lock
 *                 function")
 *  86bd446b5ceb ("net: sched: rename qdisc_destroy() to qdisc_put()")
 *  3a7d0d07a386 ("net: sched: extend Qdisc with rcu")
 *  9d7e82cec35c ("net: sched: add helper function to take reference to Qdisc")
 *  e368fdb61d8e ("net: sched: use Qdisc rcu API instead of relying on rtnl
 *                 lock")
 *
 *  SLE12-SP3 commits:
 *  none yet
 *
 *  SLE12-SP4 and SLE15 commits:
 *  3537686b788ac2c916bcc3f49f3fb4d0b589716b
 *  e3693cb4f906801a9030e9646c666466a6ed4ce7
 *  6b54018d29946eceaf26f98351def972012fcf6f
 *  9d1e163478cf17bdd1eb533a8869bf5a2031f3c2
 *  a22ecb0bdad05061a056497576aefa96d8bb28ab
 *  883b3bec552a91b750411201930800e1a11777b4
 *
 *  SLE12-SP5 and SLE15-SP1 commits:
 *  993a594b297502cd071c0b910ae064d08503e1d1
 *  d2b09ab65cba28927857e154a382f597ab1c5ca2
 *  3a9faa14dcc7ca4e43c273479e958092eb3ff7c9
 *  57b3c906d8daa3dfec10bbaa7770500a0de4d794
 *  e170ca2557706bc31f50bc30875c1b119bd1d50e
 *
 *  SLE15-SP2 and -SP3 commits:
 *  not affected
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from net/sched/cls_api.c */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/skbuff.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/err.h>
#include <linux/slab.h>
#include <net/net_namespace.h>
#include <net/sock.h>

/* klp-ccp: from include/net/sch_generic.h */
static bool (*kgre_tcf_destroy)(struct tcf_proto *tp, bool force);

/* New */
static inline struct Qdisc *kgrp_qdisc_refcount_inc_nz(struct Qdisc *qdisc)
{
	if (qdisc->flags & TCQ_F_BUILTIN)
		return qdisc;
	if (atomic_inc_not_zero(&qdisc->refcnt))
		return qdisc;
	return NULL;
}

/* klp-ccp: from net/sched/cls_api.c */
#include <net/netlink.h>

/* klp-ccp: from include/net/pkt_sched.h */
static struct Qdisc *(*kgre_qdisc_lookup)(struct net_device *dev, u32 handle);

/* klp-ccp: from net/sched/cls_api.c */
static const struct tcf_proto_ops *(*kgre_tcf_proto_lookup_ops)(struct nlattr *kind);

static int (*kgre_tfilter_notify)(struct net *net, struct sk_buff *oskb,
			  struct nlmsghdr *n, struct tcf_proto *tp,
			  unsigned long fh, int event);

static inline u32 tcf_auto_prio(struct tcf_proto *tp)
{
	u32 first = TC_H_MAKE(0xC0000000U, 0U);

	if (tp)
		first = tp->prio - 1;

	return first;
}

int kgrp_tc_ctl_tfilter(struct sk_buff *skb, struct nlmsghdr *n)
{
	struct net *net = sock_net(skb->sk);
	struct nlattr *tca[TCA_MAX + 1];
	struct tcmsg *t;
	u32 protocol;
	u32 prio;
	u32 nprio;
	u32 parent;
	struct net_device *dev;
	struct Qdisc  *q;
	struct tcf_proto __rcu **back;
	struct tcf_proto __rcu **chain;
	struct tcf_proto *tp;
	const struct tcf_proto_ops *tp_ops;
	const struct Qdisc_class_ops *cops;
	unsigned long cl;
	unsigned long fh;
	int err;
	int tp_created;

	if ((n->nlmsg_type != RTM_GETTFILTER) &&
	    !netlink_ns_capable(skb, net->user_ns, CAP_NET_ADMIN))
		return -EPERM;

replay:
	tp_created = 0;

	err = nlmsg_parse(n, sizeof(*t), tca, TCA_MAX, NULL);
	if (err < 0)
		return err;

	t = nlmsg_data(n);
	protocol = TC_H_MIN(t->tcm_info);
	prio = TC_H_MAJ(t->tcm_info);
	nprio = prio;
	parent = t->tcm_parent;
	cl = 0;

	if (prio == 0) {
		/* If no priority is given, user wants we allocated it. */
		if (n->nlmsg_type != RTM_NEWTFILTER ||
		    !(n->nlmsg_flags & NLM_F_CREATE))
			return -ENOENT;
		prio = TC_H_MAKE(0x80000000U, 0U);
	}

	/* Find head of filter chain. */

	/* Find link */
	dev = __dev_get_by_index(net, t->tcm_ifindex);
	if (dev == NULL)
		return -ENODEV;

	/* Find qdisc */
	if (!parent) {
		q = dev->qdisc;
		parent = q->handle;
	} else {
		q = (*kgre_qdisc_lookup)(dev, TC_H_MAJ(t->tcm_parent));
		if (q == NULL)
			return -EINVAL;
	}

	/*
	 * Fix CVE-2021-39713
	 *  +3 lines
	 */
	q = kgrp_qdisc_refcount_inc_nz(q);
	if (!q)
		return -EINVAL;

	/* Is it classful? */
	cops = q->ops->cl_ops;
	/*
	 * Fix CVE-2021-39713
	 *  -2 lines, +4 lines
	 */
	if (!cops) {
		err = -EINVAL;
		goto errout;
	}

	/*
	 * Fix CVE-2021-39713
	 *  -2 lines, +4 lines
	 */
	if (cops->tcf_chain == NULL) {
		err = -EOPNOTSUPP;
		goto errout;
	}

	/* Do we search for filter, attached to class? */
	if (TC_H_MIN(parent)) {
		cl = cops->get(q, parent);
		/*
		 * Fix CVE-2021-39713
		 *  -2 lines, +4 lines
		 */
		if (cl == 0) {
			err = -ENOENT;
			goto errout;
		}
	}

	/* And the last stroke */
	chain = cops->tcf_chain(q, cl);
	err = -EINVAL;
	if (chain == NULL)
		goto errout;

	/* Check the chain for existence of proto-tcf with this priority */
	for (back = chain;
	     (tp = rtnl_dereference(*back)) != NULL;
	     back = &tp->next) {
		if (tp->prio >= prio) {
			if (tp->prio == prio) {
				if (!nprio ||
				    (tp->protocol != protocol && protocol))
					goto errout;
			} else
				tp = NULL;
			break;
		}
	}

	if (tp == NULL) {
		/* Proto-tcf does not exist, create new one */

		if (tca[TCA_KIND] == NULL || !protocol)
			goto errout;

		err = -ENOENT;
		if (n->nlmsg_type != RTM_NEWTFILTER ||
		    !(n->nlmsg_flags & NLM_F_CREATE))
			goto errout;


		/* Create new proto tcf */

		err = -ENOBUFS;
		tp = kzalloc(sizeof(*tp), GFP_KERNEL);
		if (tp == NULL)
			goto errout;
		err = -ENOENT;
		tp_ops = (*kgre_tcf_proto_lookup_ops)(tca[TCA_KIND]);
		if (tp_ops == NULL) {
#ifdef CONFIG_MODULES
			struct nlattr *kind = tca[TCA_KIND];
			char name[IFNAMSIZ];

			if (kind != NULL &&
			    nla_strlcpy(name, kind, IFNAMSIZ) < IFNAMSIZ) {
				rtnl_unlock();
				request_module("cls_%s", name);
				rtnl_lock();
				tp_ops = (*kgre_tcf_proto_lookup_ops)(kind);
				/* We dropped the RTNL semaphore in order to
				 * perform the module load.  So, even if we
				 * succeeded in loading the module we have to
				 * replay the request.  We indicate this using
				 * -EAGAIN.
				 */
				if (tp_ops != NULL) {
					module_put(tp_ops->owner);
					err = -EAGAIN;
				}
			}
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
			kfree(tp);
			goto errout;
		}
		tp->ops = tp_ops;
		tp->protocol = protocol;
		tp->prio = nprio ? :
			       TC_H_MAJ(tcf_auto_prio(rtnl_dereference(*back)));
		tp->q = q;
		tp->classify = tp_ops->classify;
		tp->classid = parent;

		err = tp_ops->init(tp);
		if (err != 0) {
			module_put(tp_ops->owner);
			kfree(tp);
			goto errout;
		}

		tp_created = 1;

	} else if (tca[TCA_KIND] && nla_strcmp(tca[TCA_KIND], tp->ops->kind))
		goto errout;

	fh = tp->ops->get(tp, t->tcm_handle);

	if (fh == 0) {
		if (n->nlmsg_type == RTM_DELTFILTER && t->tcm_handle == 0) {
			struct tcf_proto *next = rtnl_dereference(tp->next);

			RCU_INIT_POINTER(*back, next);

			(*kgre_tfilter_notify)(net, skb, n, tp, fh, RTM_DELTFILTER);
			(*kgre_tcf_destroy)(tp, true);
			err = 0;
			goto errout;
		}

		err = -ENOENT;
		if (n->nlmsg_type != RTM_NEWTFILTER ||
		    !(n->nlmsg_flags & NLM_F_CREATE))
			goto errout;
	} else {
		switch (n->nlmsg_type) {
		case RTM_NEWTFILTER:
			err = -EEXIST;
			if (n->nlmsg_flags & NLM_F_EXCL) {
				if (tp_created)
					(*kgre_tcf_destroy)(tp, true);
				goto errout;
			}
			break;
		case RTM_DELTFILTER:
			err = tp->ops->delete(tp, fh);
			if (err == 0) {
				struct tcf_proto *next = rtnl_dereference(tp->next);

				(*kgre_tfilter_notify)(net, skb, n, tp,
					       t->tcm_handle, RTM_DELTFILTER);
				if ((*kgre_tcf_destroy)(tp, false))
					RCU_INIT_POINTER(*back, next);
			}
			goto errout;
		case RTM_GETTFILTER:
			err = (*kgre_tfilter_notify)(net, skb, n, tp, fh, RTM_NEWTFILTER);
			goto errout;
		default:
			err = -EINVAL;
			goto errout;
		}
	}

	err = tp->ops->change(net, skb, tp, cl, t->tcm_handle, tca, &fh,
			      n->nlmsg_flags & NLM_F_CREATE ? TCA_ACT_NOREPLACE : TCA_ACT_REPLACE);
	if (err == 0) {
		if (tp_created) {
			RCU_INIT_POINTER(tp->next, rtnl_dereference(*back));
			rcu_assign_pointer(*back, tp);
		}
		(*kgre_tfilter_notify)(net, skb, n, tp, fh, RTM_NEWTFILTER);
	} else {
		if (tp_created)
			(*kgre_tcf_destroy)(tp, true);
	}

errout:
	if (cl)
		cops->put(q, cl);
	/*
	 * Fix CVE-2021-39713
	 *  +2 lines
	 */
	if (q)
		qdisc_destroy(q);
	if (err == -EAGAIN)
		/* Replay the request. */
		goto replay;
	return err;
}




#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1197211.h"
#include "../kallsyms_relocs.h"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "qdisc_lookup", (void *)&kgre_qdisc_lookup },
	{ "tcf_destroy", (void *)&kgre_tcf_destroy },
	{ "tcf_proto_lookup_ops", (void *)&kgre_tcf_proto_lookup_ops },
	{ "tfilter_notify", (void *)&kgre_tfilter_notify },
};

int kgr_patch_bsc1197211_init(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}
